################################################################
# Copyright 1995 by
# The University of Chicago Biological Sciences Division, Academic Computing
# Lee Newberg
################################################################

# An object-oriented design of a viewed HTML page.  The various logical objects
# (e.g., page, para, image, horizontal rule, form input, form select, form
# textarea) will keep track of their own X widgets and other parameters (e.g.,
# cursor position, background color, scroll bars, height, width, edit/browse
# mode).

# Since Tcl is somewhat braindead, each instance of an object will be given its
# own id (e.g., "page=4", "para=3") and can store information about itself in
# the phxW3Obj array.  Note that although objects can be nested, phxW3Obj array
# does not use this information in its array indices.  Thus, although para=3
# may be contained in page=4, we use "para=3" as an array index rather than
# "page=4,para=3".

################################################################
# phxW3ObjInitialize:  Call at Phoenix startup to initialize this file.
################################################################

proc phxW3ObjInitialize {} {
    global phxW3Obj
    catch {unset phxW3Obj}
}

################################################################
#                                                              #
#                       HELPER ROUTINES                        #
#                       HELPER ROUTINES                        #
#                       HELPER ROUTINES                        #
#                                                              #
################################################################

################################################################
# phxW3ObjNewName: Gets a name for a new instance of a given type.
################################################################

proc phxW3ObjNewName {type} {
    global phxW3Obj
    if {[info exists phxW3Obj(count,$type)]} {
	set number [incr phxW3Obj(count,$type)]
    } {
	set number [set phxW3Obj(count,$type) 1]
    }
    return "$type=$number"
}

################################################################
# phxW3ObjListToPairs:  Converts {a b c d e f g h ...} to
# {{a b} {c d} {e f} {g h} ...}
################################################################

proc phxW3ObjListToPairs {list} {
    set pairs {}
    while {$list != {}} {
	lappend pairs [lrange $list 0 1]
	set list [lreplace $list 0 1]
    }
    return $pairs
}

################################################################
#                                                              #
#                           CLASSES                            #
#                           CLASSES                            #
#                           CLASSES                            #
#                                                              #
################################################################


################################################################
# Generic Class
# 
# phxW3ObjGeneric: Generic object for parameters applicable to
# all HTML contexts.
################################################################

proc phxW3ObjGeneric {{name ""} {method "list"} {etc ""}} {
    global phxW3Obj
    switch $method {
	browse {set phxW3Obj($name,mode) "browse"}
	destroy {
	    foreach param {cursor mode height width} {
		unset phxW3Obj($name,$param)
	    }
	}
	edit {set phxW3Obj($name,mode) "edit"}
	list {return "browse destroy edit list new"}
	new {
	    set phxW3Obj($name,cursor) "none"
	    set phxW3Obj($name,mode) "browse"
	    set phxW3Obj($name,height) "unlimited"
	    set phxW3Obj($name,width) "unlimited"
	}
	default {
	    error!!!
	}
    }
}

################################################################
# Page Class
# 
# Routines for handling a "page" which represents any HTML that might occur
# between <BODY> and </BODY> marks.  
################################################################

# phxW3ObjPage controls all the methods for the class Page.
# The first argument is the method invoked,
# the second argument is the page instance, and
# the third argument is a *list* of further arguments.

proc phxW3ObjPage {{name ""} {method "list"} {etc ""}} {
    global phxW3Obj
    switch $method {
	body	{return [$phxW3Obj($name,body) inner]}
	destroy {
	    foreach child $phxW3Obj($name,children) {
		$child destroy
	    }
	    unset phxW3Obj($name,children)

	    foreach child {body hypertext spacer scrollbar outer} {
		$phxW3Obj($name,$child) destroy
		unset phxW3Obj($name,$child)
	    }

	    phxW3ObjGeneric $name destroy $etc
	    rename $name ""
	}
	list	{return [concat "body destroy new outer" \
			[phxW3ObjGeneric $name list $etc]]}
	new	{return [phxW3ObjPageNew] $etc}
	outer	{return [$phxW3Obj($name,outer) outer]}
	default {phxW3ObjGeneric $name $method $etc}
    }
}

# phxW3ObjPageNew is the constructor for instances of the Page class.  A page
# is implemented as a frame which contains a scrollbar, an empty frame, and a
# hypertext widget.  A frame is the sole child of the hypertext widget.
# 
# Justification: Unfortunately the contents of a hypertext widget cannot be
# changed without a complete reload of its contents.  Also, frame widgets do
# not support scrollbars.  With this setup we get both a scrollbar and the
# freedom to pack (and repack) inside the frame widget.

proc phxW3ObjPageNew {etc} {
    global phxW3Obj
    set name [phxW3ObjNewName page]

    # Initialize features common among all objects.
    phxW3ObjGeneric $name new $etc

    # The "outer" frame contains the "scrollbar", a "spacer", and an
    # "hypertext" widget.  This last contains only a "body" frame.
    set phxW3Obj($name,outer) [phxW3ObjFrame new $etc]
    set phxW3Obj($name,scrollbar) [phxW3ObjScrollbar new $etc]
    set phxW3Obj($name,spacer) [phxW3ObjFrame new $etc]
    set phxW3Obj($name,hypertext) [phxW3ObjHypertext new $etc]
    set phxW3Obj($name,body) [phxW3ObjFrame new $etc]
    set phxW3Obj($name,children) {}

    # Configure and pack the scrollbar
    $phxW3Obj($name,scrollbar) configure -orient vertical \
	-borderwidth 1 -relief sunken
    $phxW3Obj($name,scrollbar) yviewwidget $phxW3Obj($name,hypertext)
    pack [$phxW3Obj($name,scrollbar) outer] \
	-in [$phxW3Obj($name,outer) inner] \
	-side left -fill y

    # Configure and pack the spacer
    $phxW3Obj($name,spacer) configure -width 4 -relief flat
    pack [$phxW3Obj($name,spacer) outer] \
	-in [$phxW3Obj($name,outer) inner] \
	-side left -fill y

    # Configure and pack the hypertext widget
    $phxW3Obj($name,hypertext) yscrollwidget $phxW3Obj($name,scrollbar)
    pack [$phxW3Obj($name,hypertext) outer] \
	-in [$phxW3Obj($name,outer) inner] \
	-side top -fill both -expand yes
    # Worry about keyboard bindings!!!

    # Configure and pack the body
    $phxW3Obj($name,hypertext) configure -text \
"%%[$phxW3Obj($name,hypertext) inner] append [$phxW3Obj($name,body) outer]
[$phxW3Obj($name,hypertext) inner] childconfig [$phxW3Obj($name,body) outer] -fill both -relwidth 1
%%
"

    # Define a function that handles this instance's methods
    proc $name {args} \
	"phxW3ObjPage $name \[lindex \$args 0\] \[lreplace \$args 0 0\]"

    return $name
}

################################################################
# Hypertext Class
################################################################

proc phxW3ObjHypertext {{name ""} {method "list"} {etc ""}} {
    global phxW3Obj
    switch $method {
	configure {
	    return [eval [linsert $etc 0 $phxW3Obj($name,widget) $method]]
	}
	destroy {
	    destroy $phxW3Obj($name,widget)
	    unset $phxW3Obj($name,widget)
	    phxW3ObjGeneric $name destroy $etc
	    rename $name ""
	}
	inner	-
	outer	{return $phxW3Obj($name,widget)}
	list	{return [concat "configure destroy inner new outer" \
			[phxW3ObjGeneric $name list $etc]]}
	new	{return [phxW3ObjHypertextNew] $etc}
	yscrollwidget {
	    $phxW3Obj($name,widget) configure -yscrollcommand \
		"[[lindex $etc 0] outer] set"
	}
	default {phxW3ObjGeneric $name $method $etc}
    }
}

proc phxW3ObjHypertextNew {etc} {
    global phxW3Obj
    set name [phxW3ObjNewName hypertext]

    # Initialize features common among all objects.
    phxW3ObjGeneric $name new $etc

    # Create and record the hypertext widget
    set phxW3Obj($name,widget) [blt_htext .$name]

    # Define a function that handles this instance's methods
    proc $name {args} \
	"phxW3ObjHypertext $name \[lindex \$args 0\] \[lreplace \$args 0 0\]"

    return $name
}

################################################################
# Frame Class
################################################################

proc phxW3ObjFrame {{name ""} {method "list"} {etc ""}} {
    global phxW3Obj
    switch $method {
	configure {
	    return [eval [linsert $etc 0 $phxW3Obj($name,widget) $method]]
	}
	destroy {
	    destroy $phxW3Obj($name,widget)
	    unset $phxW3Obj($name,widget)
	    phxW3ObjGeneric $name destroy $etc
	    rename $name ""
	}
	list	{return [concat "configure destroy inner new outer" \
			[phxW3ObjGeneric $name list $etc]]}
	new	{return [phxW3ObjFrameNew] $etc}
	inner	-
	outer	{return $phxW3Obj($name,widget)}
	default {phxW3ObjGeneric $name $method $etc}
    }
}

proc phxW3ObjFrameNew {etc} {
    global phxW3Obj
    set name [phxW3ObjNewName frame]

    # Initialize features common among all objects.
    phxW3ObjGeneric $name new $etc

    # Create and record the Frame widget
    set phxW3Obj($name,widget) [frame .$name]

    # Define a function that handles this instance's methods
    proc $name {args} \
	"phxW3ObjFrame $name \[lindex \$args 0\] \[lreplace \$args 0 0\]"

    return $name
}

################################################################
# Scrollbar Class
################################################################

proc phxW3ObjScrollbar {{name ""} {method "list"} {etc ""}} {
    global phxW3Obj

    switch $method {
	configure {
	    return [eval [linsert $etc 0 $phxW3Obj($name,widget) $method]]
	}
	destroy {
	    destroy $phxW3Obj($name,widget)
	    unset $phxW3Obj($name,widget)
	    phxW3ObjGeneric $name destroy $etc
	    rename $name ""
	}
	list	{return [concat "configure destroy new outer yviewcommand" \
			[phxW3ObjGeneric $name list $etc]]}
	new	{return [phxW3ObjScrollbarNew] $etc}
	outer	{return $phxW3Obj($name,widget)}
	yviewwidget {
	    $phxW3Obj($name,widget) configure -command [[lindex $etc 0] outer]
	}
	default {phxW3ObjGeneric $name $method $etc}
    }
}

proc phxW3ObjScrollbarNew {etc} {
    global phxW3Obj
    set name [phxW3ObjNewName scrollbar]

    # Initialize features common among all objects.
    phxW3ObjGeneric $name new $etc

    # Create and record the Scrollbar widget
    set phxW3Obj($name,widget) [scrollbar .$name]

    # Define a function that handles this instance's methods
    proc $name {args} \
	"phxW3ObjScrollbar $name \[lindex \$args 0\] \[lreplace \$args 0 0\]"

    return $name
}

